// Copyright 2021 - 2022 Matrix Origin
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package function

import (
	"math"

	"github.com/matrixorigin/matrixone/pkg/container/types"
	"github.com/matrixorigin/matrixone/pkg/container/vector"
	"github.com/matrixorigin/matrixone/pkg/pb/plan"
	"github.com/matrixorigin/matrixone/pkg/sql/plan/function/operator"
	"github.com/matrixorigin/matrixone/pkg/vm/process"
)

func initOperators() {
	var err error

	for fid, fs := range operators {
		err = appendFunction(fid, fs)
		if err != nil {
			panic(err)
		}
	}
}

// operators contains the operator function indexed by function id.
var operators = map[int]Functions{
	// is null operator
	ISNULL: {
		Id: ISNULL,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_uint8,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNull[uint8],
			},
			{
				Index:  1,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_uint16,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNull[uint16],
			},
			{
				Index:  2,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_uint32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNull[uint32],
			},
			{
				Index:  3,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_uint64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNull[uint64],
			},
			{
				Index:  4,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_int8,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNull[int8],
			},
			{
				Index:  5,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_int16,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNull[int16],
			},
			{
				Index:  6,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_int32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNull[int32],
			},
			{
				Index:  7,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_int64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNull[int64],
			},
			{
				Index:  8,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_float32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNull[float32],
			},
			{
				Index:  9,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_float64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNull[float64],
			},
			{
				Index:  10,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_decimal64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNull[types.Decimal64],
			},
			{
				Index:  11,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_decimal128,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNull[types.Decimal128],
			},
			{
				Index:  12,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_varchar,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsStringNull,
			},
			{
				Index:  13,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_char,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsStringNull,
			},
			{
				Index:  14,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_date,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNull[types.Date],
			},
			{
				Index:  15,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_datetime,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNull[types.Datetime],
			},
			{
				Index:  16,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_bool,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNull[bool],
			},
			{
				Index:  17,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_blob,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsStringNull,
			},
		},
	},

	ISNOTNULL: {
		Id: ISNOTNULL,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_uint8,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNotNull[uint8],
			},
			{
				Index:  1,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_uint16,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNotNull[uint16],
			},
			{
				Index:  2,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_uint32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNotNull[uint32],
			},
			{
				Index:  3,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_uint64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNotNull[uint64],
			},
			{
				Index:  4,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_int8,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNotNull[int8],
			},
			{
				Index:  5,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_int16,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNotNull[int16],
			},
			{
				Index:  6,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_int32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNotNull[int32],
			},
			{
				Index:  7,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_int64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNotNull[int64],
			},
			{
				Index:  8,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_float32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNotNull[float32],
			},
			{
				Index:  9,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_float64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNotNull[float64],
			},
			{
				Index:  10,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_decimal64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNotNull[types.Decimal64],
			},
			{
				Index:  11,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_decimal128,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNotNull[types.Decimal128],
			},
			{
				Index:  12,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_varchar,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsStringNotNull,
			},
			{
				Index:  13,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_char,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsStringNotNull,
			},
			{
				Index:  14,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_date,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNotNull[types.Date],
			},
			{
				Index:  15,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_datetime,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNotNull[types.Datetime],
			},
			{
				Index:  16,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_bool,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNotNull[bool],
			},
			{
				Index:  17,
				Flag:   plan.Function_STRICT,
				Layout: IS_NULL_EXPRESSION,
				Args: []types.T{
					types.T_blob,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsStringNotNull,
			},
		},
	},
	// comparison operator
	IS: {
		Id: IS,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_bool,
					types.T_bool,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.Is,
			},
		},
	},

	OP_BIT_XOR: {
		Id: OP_BIT_XOR,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint8,
					types.T_uint8,
				},
				ReturnTyp: types.T_uint8,
				Fn:        operator.OpBitXorFun[uint8],
			},
			{
				Index:  1,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int8,
					types.T_int8,
				},
				ReturnTyp: types.T_int8,
				Fn:        operator.OpBitXorFun[int8],
			},
			{
				Index:  2,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint16,
					types.T_uint16,
				},
				ReturnTyp: types.T_uint16,
				Fn:        operator.OpBitXorFun[uint16],
			},
			{
				Index:  3,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int16,
					types.T_int16,
				},
				ReturnTyp: types.T_int16,
				Fn:        operator.OpBitXorFun[int16],
			},
			{
				Index:  4,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint32,
					types.T_uint32,
				},
				ReturnTyp: types.T_uint32,
				Fn:        operator.OpBitXorFun[uint32],
			},
			{
				Index:  5,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int32,
					types.T_int32,
				},
				ReturnTyp: types.T_int32,
				Fn:        operator.OpBitXorFun[int32],
			},
			{
				Index:  6,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint64,
					types.T_uint64,
				},
				ReturnTyp: types.T_uint64,
				Fn:        operator.OpBitXorFun[uint64],
			},
			{
				Index:  7,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int64,
					types.T_int64,
				},
				ReturnTyp: types.T_int64,
				Fn:        operator.OpBitXorFun[int64],
			},
		},
	},

	OP_BIT_OR: {
		Id: OP_BIT_OR,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint8,
					types.T_uint8,
				},
				ReturnTyp: types.T_uint8,
				Fn:        operator.OpBitOrFun[uint8],
			},
			{
				Index:  1,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int8,
					types.T_int8,
				},
				ReturnTyp: types.T_int8,
				Fn:        operator.OpBitOrFun[int8],
			},
			{
				Index:  2,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint16,
					types.T_uint16,
				},
				ReturnTyp: types.T_uint16,
				Fn:        operator.OpBitOrFun[uint16],
			},
			{
				Index:  3,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int16,
					types.T_int16,
				},
				ReturnTyp: types.T_int16,
				Fn:        operator.OpBitOrFun[int16],
			},
			{
				Index:  4,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint32,
					types.T_uint32,
				},
				ReturnTyp: types.T_uint32,
				Fn:        operator.OpBitOrFun[uint32],
			},
			{
				Index:  5,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int32,
					types.T_int32,
				},
				ReturnTyp: types.T_int32,
				Fn:        operator.OpBitOrFun[int32],
			},
			{
				Index:  6,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint64,
					types.T_uint64,
				},
				ReturnTyp: types.T_uint64,
				Fn:        operator.OpBitOrFun[uint64],
			},
			{
				Index:  7,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int64,
					types.T_int64,
				},
				ReturnTyp: types.T_int64,
				Fn:        operator.OpBitOrFun[int64],
			},
		},
	},

	OP_BIT_AND: {
		Id: OP_BIT_AND,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint8,
					types.T_uint8,
				},
				ReturnTyp: types.T_uint8,
				Fn:        operator.OpBitAndFun[uint8],
			},
			{
				Index:  1,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int8,
					types.T_int8,
				},
				ReturnTyp: types.T_int8,
				Fn:        operator.OpBitAndFun[int8],
			},
			{
				Index:  2,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint16,
					types.T_uint16,
				},
				ReturnTyp: types.T_uint16,
				Fn:        operator.OpBitAndFun[uint16],
			},
			{
				Index:  3,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int16,
					types.T_int16,
				},
				ReturnTyp: types.T_int16,
				Fn:        operator.OpBitAndFun[int16],
			},
			{
				Index:  4,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint32,
					types.T_uint32,
				},
				ReturnTyp: types.T_uint32,
				Fn:        operator.OpBitAndFun[uint32],
			},
			{
				Index:  5,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int32,
					types.T_int32,
				},
				ReturnTyp: types.T_int32,
				Fn:        operator.OpBitAndFun[int32],
			},
			{
				Index:  6,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint64,
					types.T_uint64,
				},
				ReturnTyp: types.T_uint64,
				Fn:        operator.OpBitAndFun[uint64],
			},
			{
				Index:  7,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int64,
					types.T_int64,
				},
				ReturnTyp: types.T_int64,
				Fn:        operator.OpBitAndFun[int64],
			},
		},
	},

	OP_BIT_SHIFT_RIGHT: {
		Id: OP_BIT_SHIFT_RIGHT,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint8,
					types.T_uint8,
				},
				ReturnTyp: types.T_uint8,
				Fn:        operator.OpBitRightShiftFun[uint8],
			},
			{
				Index:  1,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int8,
					types.T_int8,
				},
				ReturnTyp: types.T_int8,
				Fn:        operator.OpBitRightShiftFun[int8],
			},
			{
				Index:  2,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint16,
					types.T_uint16,
				},
				ReturnTyp: types.T_uint16,
				Fn:        operator.OpBitRightShiftFun[uint16],
			},
			{
				Index:  3,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int16,
					types.T_int16,
				},
				ReturnTyp: types.T_int16,
				Fn:        operator.OpBitRightShiftFun[int16],
			},
			{
				Index:  4,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint32,
					types.T_uint32,
				},
				ReturnTyp: types.T_uint32,
				Fn:        operator.OpBitRightShiftFun[uint32],
			},
			{
				Index:  5,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int32,
					types.T_int32,
				},
				ReturnTyp: types.T_int32,
				Fn:        operator.OpBitRightShiftFun[int32],
			},
			{
				Index:  6,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint64,
					types.T_uint64,
				},
				ReturnTyp: types.T_uint64,
				Fn:        operator.OpBitRightShiftFun[uint64],
			},
			{
				Index:  7,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int64,
					types.T_int64,
				},
				ReturnTyp: types.T_int64,
				Fn:        operator.OpBitRightShiftFun[int64],
			},
		},
	},

	OP_BIT_SHIFT_LEFT: {
		Id: OP_BIT_SHIFT_LEFT,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint8,
					types.T_uint8,
				},
				ReturnTyp: types.T_uint8,
				Fn:        operator.OpBitLeftShiftFun[uint8],
			},
			{
				Index:  1,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int8,
					types.T_int8,
				},
				ReturnTyp: types.T_int8,
				Fn:        operator.OpBitLeftShiftFun[int8],
			},
			{
				Index:  2,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint16,
					types.T_uint16,
				},
				ReturnTyp: types.T_uint16,
				Fn:        operator.OpBitLeftShiftFun[uint16],
			},
			{
				Index:  3,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int16,
					types.T_int16,
				},
				ReturnTyp: types.T_int16,
				Fn:        operator.OpBitLeftShiftFun[int16],
			},
			{
				Index:  4,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint32,
					types.T_uint32,
				},
				ReturnTyp: types.T_uint32,
				Fn:        operator.OpBitLeftShiftFun[uint32],
			},
			{
				Index:  5,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int32,
					types.T_int32,
				},
				ReturnTyp: types.T_int32,
				Fn:        operator.OpBitLeftShiftFun[int32],
			},
			{
				Index:  6,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint64,
					types.T_uint64,
				},
				ReturnTyp: types.T_uint64,
				Fn:        operator.OpBitLeftShiftFun[uint64],
			},
			{
				Index:  7,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int64,
					types.T_int64,
				},
				ReturnTyp: types.T_int64,
				Fn:        operator.OpBitLeftShiftFun[int64],
			},
		},
	},

	ISNOT: {
		Id: ISNOT,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_bool,
					types.T_bool,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.IsNot,
			},
		},
	},

	EQUAL: {
		Id:          EQUAL,
		TypeCheckFn: GeneralBinaryOperatorTypeCheckFn1,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint8,
					types.T_uint8,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqGeneral[uint8],
			},
			{
				Index:  1,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint16,
					types.T_uint16,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqGeneral[uint16],
			},
			{
				Index:  2,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint32,
					types.T_uint32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqGeneral[uint32],
			},
			{
				Index:  3,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint64,
					types.T_uint64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqGeneral[uint64],
			},
			{
				Index:  4,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int8,
					types.T_int8,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqGeneral[int8],
			},
			{
				Index:  5,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int16,
					types.T_int16,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqGeneral[int16],
			},
			{
				Index:  6,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int32,
					types.T_int32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqGeneral[int32],
			},
			{
				Index:  7,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int64,
					types.T_int64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqGeneral[int64],
			},
			{
				Index:  8,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_float32,
					types.T_float32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqGeneral[float32],
			},
			{
				Index:  9,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_float64,
					types.T_float64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqGeneral[float64],
			},
			{
				Index:  10,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_decimal64,
					types.T_decimal64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqDecimal64,
			},
			{
				Index:  11,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_decimal128,
					types.T_decimal128,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqDecimal128,
			},
			{
				Index:  12,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_varchar,
					types.T_varchar,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqString,
			},
			{
				Index:  13,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_char,
					types.T_char,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqString,
			},
			{
				Index:  14,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_date,
					types.T_date,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqGeneral[types.Date],
			},
			{
				Index:  15,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_datetime,
					types.T_datetime,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqGeneral[types.Datetime],
			},
			{
				Index:  16,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_bool,
					types.T_bool,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqGeneral[bool],
			},
			{
				Index:  17,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_timestamp,
					types.T_timestamp,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqGeneral[types.Timestamp],
			},
			{
				Index:  18,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_blob,
					types.T_blob,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.EqString,
			},
		},
	},

	GREAT_THAN: {
		Id:          EQUAL,
		TypeCheckFn: GeneralBinaryOperatorTypeCheckFn1,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint8,
					types.T_uint8,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtGeneral[uint8],
			},
			{
				Index:  1,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint16,
					types.T_uint16,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtGeneral[uint16],
			},
			{
				Index:  2,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint32,
					types.T_uint32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtGeneral[uint32],
			},
			{
				Index:  3,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint64,
					types.T_uint64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtGeneral[uint64],
			},
			{
				Index:  4,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int8,
					types.T_int8,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtGeneral[int8],
			},
			{
				Index:  5,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int16,
					types.T_int16,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtGeneral[int16],
			},
			{
				Index:  6,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int32,
					types.T_int32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtGeneral[int32],
			},
			{
				Index:  7,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int64,
					types.T_int64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtGeneral[int64],
			},
			{
				Index:  8,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_float32,
					types.T_float32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtGeneral[float32],
			},
			{
				Index:  9,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_float64,
					types.T_float64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtGeneral[float64],
			},
			{
				Index:  10,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_decimal64,
					types.T_decimal64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtDecimal64,
			},
			{
				Index:  11,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_decimal128,
					types.T_decimal128,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtDecimal128,
			},
			{
				Index:  12,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_varchar,
					types.T_varchar,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtString,
			},
			{
				Index:  13,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_char,
					types.T_char,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtString,
			},
			{
				Index:  14,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_date,
					types.T_date,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtGeneral[types.Date],
			},
			{
				Index:  15,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_datetime,
					types.T_datetime,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtGeneral[types.Datetime],
			},
			{
				Index:  16,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_bool,
					types.T_bool,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtGeneral[bool],
			},
			{
				Index:  17,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_timestamp,
					types.T_timestamp,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtGeneral[types.Timestamp],
			},
			{
				Index:  18,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_blob,
					types.T_blob,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GtString,
			},
		},
	},

	GREAT_EQUAL: {
		Id:          GREAT_EQUAL,
		TypeCheckFn: GeneralBinaryOperatorTypeCheckFn1,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint8,
					types.T_uint8,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeGeneral[uint8],
			},
			{
				Index:  1,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint16,
					types.T_uint16,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeGeneral[uint16],
			},
			{
				Index:  2,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint32,
					types.T_uint32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeGeneral[uint32],
			},
			{
				Index:  3,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint64,
					types.T_uint64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeGeneral[uint64],
			},
			{
				Index:  4,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int8,
					types.T_int8,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeGeneral[int8],
			},
			{
				Index:  5,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int16,
					types.T_int16,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeGeneral[int16],
			},
			{
				Index:  6,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int32,
					types.T_int32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeGeneral[int32],
			},
			{
				Index:  7,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int64,
					types.T_int64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeGeneral[int64],
			},
			{
				Index:  8,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_float32,
					types.T_float32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeGeneral[float32],
			},
			{
				Index:  9,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_float64,
					types.T_float64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeGeneral[float64],
			},
			{
				Index:  10,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_decimal64,
					types.T_decimal64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeDecimal64,
			},
			{
				Index:  11,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_decimal128,
					types.T_decimal128,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeDecimal128,
			},
			{
				Index:  12,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_varchar,
					types.T_varchar,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeString,
			},
			{
				Index:  13,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_char,
					types.T_char,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeString,
			},
			{
				Index:  14,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_date,
					types.T_date,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeGeneral[types.Date],
			},
			{
				Index:  15,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_datetime,
					types.T_datetime,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeGeneral[types.Datetime],
			},
			{
				Index:  16,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_bool,
					types.T_bool,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeGeneral[bool],
			},
			{
				Index:  17,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_timestamp,
					types.T_timestamp,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeGeneral[types.Timestamp],
			},
			{
				Index:  18,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_blob,
					types.T_blob,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.GeString,
			},
		},
	},

	LESS_THAN: {
		Id:          LESS_THAN,
		TypeCheckFn: GeneralBinaryOperatorTypeCheckFn1,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint8,
					types.T_uint8,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtGeneral[uint8],
			},
			{
				Index:  1,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint16,
					types.T_uint16,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtGeneral[uint16],
			},
			{
				Index:  2,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint32,
					types.T_uint32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtGeneral[uint32],
			},
			{
				Index:  3,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint64,
					types.T_uint64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtGeneral[uint64],
			},
			{
				Index:  4,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int8,
					types.T_int8,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtGeneral[int8],
			},
			{
				Index:  5,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int16,
					types.T_int16,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtGeneral[int16],
			},
			{
				Index:  6,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int32,
					types.T_int32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtGeneral[int32],
			},
			{
				Index:  7,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int64,
					types.T_int64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtGeneral[int64],
			},
			{
				Index:  8,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_float32,
					types.T_float32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtGeneral[float32],
			},
			{
				Index:  9,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_float64,
					types.T_float64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtGeneral[float64],
			},
			{
				Index:  10,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_decimal64,
					types.T_decimal64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtDecimal64,
			},
			{
				Index:  11,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_decimal128,
					types.T_decimal128,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtDecimal128,
			},
			{
				Index:  12,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_varchar,
					types.T_varchar,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtString,
			},
			{
				Index:  13,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_char,
					types.T_char,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtString,
			},
			{
				Index:  14,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_date,
					types.T_date,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtGeneral[types.Date],
			},
			{
				Index:  15,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_datetime,
					types.T_datetime,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtGeneral[types.Datetime],
			},
			{
				Index:  16,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_bool,
					types.T_bool,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtGeneral[bool],
			},
			{
				Index:  17,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_timestamp,
					types.T_timestamp,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtGeneral[types.Timestamp],
			},
			{
				Index:  18,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_blob,
					types.T_blob,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LtString,
			},
		},
	},

	LESS_EQUAL: {
		Id:          LESS_EQUAL,
		TypeCheckFn: GeneralBinaryOperatorTypeCheckFn1,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint8,
					types.T_uint8,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeGeneral[uint8],
			},
			{
				Index:  1,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint16,
					types.T_uint16,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeGeneral[uint16],
			},
			{
				Index:  2,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint32,
					types.T_uint32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeGeneral[uint32],
			},
			{
				Index:  3,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint64,
					types.T_uint64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeGeneral[uint64],
			},
			{
				Index:  4,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int8,
					types.T_int8,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeGeneral[int8],
			},
			{
				Index:  5,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int16,
					types.T_int16,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeGeneral[int16],
			},
			{
				Index:  6,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int32,
					types.T_int32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeGeneral[int32],
			},
			{
				Index:  7,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int64,
					types.T_int64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeGeneral[int64],
			},
			{
				Index:  8,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_float32,
					types.T_float32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeGeneral[float32],
			},
			{
				Index:  9,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_float64,
					types.T_float64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeGeneral[float64],
			},
			{
				Index:  10,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_decimal64,
					types.T_decimal64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeDecimal64,
			},
			{
				Index:  11,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_decimal128,
					types.T_decimal128,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeDecimal128,
			},
			{
				Index:  12,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_varchar,
					types.T_varchar,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeString,
			},
			{
				Index:  13,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_char,
					types.T_char,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeString,
			},
			{
				Index:  14,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_date,
					types.T_date,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeGeneral[types.Date],
			},
			{
				Index:  15,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_datetime,
					types.T_datetime,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeGeneral[types.Datetime],
			},
			{
				Index:  16,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_bool,
					types.T_bool,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeGeneral[bool],
			},
			{
				Index:  17,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_timestamp,
					types.T_timestamp,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeGeneral[types.Timestamp],
			},
			{
				Index:  18,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_blob,
					types.T_blob,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LeString,
			},
		},
	},

	NOT_EQUAL: {
		Id:          NOT_EQUAL,
		TypeCheckFn: GeneralBinaryOperatorTypeCheckFn1,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint8,
					types.T_uint8,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeGeneral[uint8],
			},
			{
				Index:  1,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint16,
					types.T_uint16,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeGeneral[uint16],
			},
			{
				Index:  2,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint32,
					types.T_uint32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeGeneral[uint16],
			},
			{
				Index:  3,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_uint64,
					types.T_uint64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeGeneral[uint64],
			},
			{
				Index:  4,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int8,
					types.T_int8,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeGeneral[int8],
			},
			{
				Index:  5,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int16,
					types.T_int16,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeGeneral[int16],
			},
			{
				Index:  6,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int32,
					types.T_int32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeGeneral[int32],
			},
			{
				Index:  7,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_int64,
					types.T_int64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeGeneral[int64],
			},
			{
				Index:  8,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_float32,
					types.T_float32,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeGeneral[float32],
			},
			{
				Index:  9,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_float64,
					types.T_float64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeGeneral[float64],
			},
			{
				Index:  10,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_decimal64,
					types.T_decimal64,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeDecimal64,
			},
			{
				Index:  11,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_decimal128,
					types.T_decimal128,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeDecimal128,
			},
			{
				Index:  12,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_varchar,
					types.T_varchar,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeString,
			},
			{
				Index:  13,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_char,
					types.T_char,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeString,
			},
			{
				Index:  14,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_date,
					types.T_date,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeGeneral[types.Date],
			},
			{
				Index:  15,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_datetime,
					types.T_datetime,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeGeneral[types.Datetime],
			},
			{
				Index:  16,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_bool,
					types.T_bool,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeGeneral[bool],
			},
			{
				Index:  17,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_timestamp,
					types.T_timestamp,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeGeneral[types.Timestamp],
			},
			{
				Index:  18,
				Flag:   plan.Function_STRICT,
				Layout: COMPARISON_OPERATOR,
				Args: []types.T{
					types.T_blob,
					types.T_blob,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.NeString,
			},
		},
	},

	LIKE: {
		Id: LIKE,
		TypeCheckFn: func(_ []Function, inputs []types.T) (overloadIndex int32, _ []types.T) {
			if len(inputs) != 2 {
				return wrongFunctionParameters, nil
			}
			typ1, typ2 := inputs[0], inputs[1]
			if typ1 != types.T_char && typ1 != types.T_varchar && typ1 != types.T_blob {
				return wrongFunctionParameters, nil
			}
			if typ2 != types.T_char && typ2 != types.T_varchar && typ2 != types.T_blob {
				return wrongFunctionParameters, nil
			}
			return 0, nil
		},
		Overloads: []Function{
			{
				Index:     0,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_LOGICAL_OPERATOR,
				Args:      nil,
				ReturnTyp: types.T_bool,
				Fn:        operator.Like,
			},
		},
	},

	BETWEEN: {
		Id:          BETWEEN,
		TypeCheckFn: GeneralBinaryOperatorTypeCheckFn1,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: BETWEEN_AND_EXPRESSION,
				Args: []types.T{
					types.T_uint8,
					types.T_uint8,
				},
				ReturnTyp: types.T_bool,
				Fn:        nil,
			},
			{
				Index:  1,
				Flag:   plan.Function_STRICT,
				Layout: BETWEEN_AND_EXPRESSION,
				Args: []types.T{
					types.T_uint16,
					types.T_uint16,
				},
				ReturnTyp: types.T_bool,
				Fn:        nil,
			},
			{
				Index:  2,
				Flag:   plan.Function_STRICT,
				Layout: BETWEEN_AND_EXPRESSION,
				Args: []types.T{
					types.T_uint32,
					types.T_uint32,
				},
				ReturnTyp: types.T_bool,
				Fn:        nil,
			},
			{
				Index:  3,
				Flag:   plan.Function_STRICT,
				Layout: BETWEEN_AND_EXPRESSION,
				Args: []types.T{
					types.T_uint64,
					types.T_uint64,
				},
				ReturnTyp: types.T_bool,
				Fn:        nil,
			},
			{
				Index:  4,
				Flag:   plan.Function_STRICT,
				Layout: BETWEEN_AND_EXPRESSION,
				Args: []types.T{
					types.T_int8,
					types.T_int8,
				},
				ReturnTyp: types.T_bool,
				Fn:        nil,
			},
			{
				Index:  5,
				Flag:   plan.Function_STRICT,
				Layout: BETWEEN_AND_EXPRESSION,
				Args: []types.T{
					types.T_int16,
					types.T_int16,
				},
				ReturnTyp: types.T_bool,
				Fn:        nil,
			},
			{
				Index:  6,
				Flag:   plan.Function_STRICT,
				Layout: BETWEEN_AND_EXPRESSION,
				Args: []types.T{
					types.T_int32,
					types.T_int32,
				},
				ReturnTyp: types.T_bool,
				Fn:        nil,
			},
			{
				Index:  7,
				Flag:   plan.Function_STRICT,
				Layout: BETWEEN_AND_EXPRESSION,
				Args: []types.T{
					types.T_int64,
					types.T_int64,
				},
				ReturnTyp: types.T_bool,
				Fn:        nil,
			},
			{
				Index:  8,
				Flag:   plan.Function_STRICT,
				Layout: BETWEEN_AND_EXPRESSION,
				Args: []types.T{
					types.T_float32,
					types.T_float32,
				},
				ReturnTyp: types.T_bool,
				Fn:        nil,
			},
			{
				Index:  9,
				Flag:   plan.Function_STRICT,
				Layout: BETWEEN_AND_EXPRESSION,
				Args: []types.T{
					types.T_float64,
					types.T_float64,
				},
				ReturnTyp: types.T_bool,
				Fn:        nil,
			},
			{
				Index:  10,
				Flag:   plan.Function_STRICT,
				Layout: BETWEEN_AND_EXPRESSION,
				Args: []types.T{
					types.T_decimal64,
					types.T_decimal64,
				},
				ReturnTyp: types.T_bool,
				Fn:        nil,
			},
			{
				Index:  11,
				Flag:   plan.Function_STRICT,
				Layout: BETWEEN_AND_EXPRESSION,
				Args: []types.T{
					types.T_decimal128,
					types.T_decimal128,
				},
				ReturnTyp: types.T_bool,
				Fn:        nil,
			},
		},
	},

	IN: {
		Id: IN,
		TypeCheckFn: func(_ []Function, inputs []types.T) (overloadIndex int32, _ []types.T) {
			if len(inputs) == 2 && inputs[1] == types.T_tuple {
				return 0, nil
			}
			return wrongFunctionParameters, nil
		},
		Overloads: []Function{
			{
				Index:     0,
				Flag:      plan.Function_STRICT,
				Layout:    IN_PREDICATE,
				ReturnTyp: types.T_bool,
			},
		},
	},

	EXISTS: {
		Id: EXISTS,
		TypeCheckFn: func(_ []Function, inputs []types.T) (overloadIndex int32, _ []types.T) {
			if len(inputs) == 1 {
				return 0, nil
			}
			return wrongFunctionParameters, nil
		},
		Overloads: []Function{
			{
				Index:     0,
				Flag:      plan.Function_STRICT,
				Layout:    EXISTS_ANY_PREDICATE,
				ReturnTyp: types.T_bool,
			},
		},
	},

	// logic operator
	AND: {
		Id: AND,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: BINARY_LOGICAL_OPERATOR,
				Args: []types.T{
					types.T_bool,
					types.T_bool,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LogicAnd,
			},
		},
	},

	OR: {
		Id: OR,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: BINARY_LOGICAL_OPERATOR,
				Args: []types.T{
					types.T_bool,
					types.T_bool,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LogicOr,
			},
		},
	},

	XOR: {
		Id: XOR,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: BINARY_LOGICAL_OPERATOR,
				Args: []types.T{
					types.T_bool,
					types.T_bool,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LogicXor,
			},
		},
	},

	NOT: {
		Id: NOT,
		Overloads: []Function{
			{
				Index:  0,
				Flag:   plan.Function_STRICT,
				Layout: UNARY_LOGICAL_OPERATOR,
				Args: []types.T{
					types.T_bool,
				},
				ReturnTyp: types.T_bool,
				Fn:        operator.LogicNot,
			},
		},
	},

	// arithmetic operator
	PLUS: {
		Id:          PLUS,
		TypeCheckFn: GeneralBinaryOperatorTypeCheckFn1,
		Overloads: []Function{
			{
				Index:     0,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint8, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.PlusUint[uint8],
			},
			{
				Index:     1,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint16, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.PlusUint[uint16],
			},
			{
				Index:     2,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint32, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.PlusUint[uint32],
			},
			{
				Index:     3,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint64, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.PlusUint[uint64],
			},
			{
				Index:     4,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int8, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.PlusInt[int8],
			},
			{
				Index:     5,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int16, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.PlusInt[int16],
			},
			{
				Index:     6,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int32, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.PlusInt[int32],
			},
			{
				Index:     7,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int64, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.PlusInt[int64],
			},
			{
				Index:     8,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_float32, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.PlusFloat[float32],
			},
			{
				Index:     9,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_float64, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.PlusFloat[float64],
			},
			{
				Index:     10,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_decimal64, types.T_decimal64},
				ReturnTyp: types.T_decimal64,
				Fn:        operator.PlusDecimal64,
			},
			{
				Index:     11,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_decimal128, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.PlusDecimal128,
			},
			{
				Index:     12,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_date, types.T_interval},
				ReturnTyp: types.T_date,
				Fn:        nil,
			},
		},
	},

	MINUS: {
		Id:          MINUS,
		TypeCheckFn: GeneralBinaryOperatorTypeCheckFn1,
		Overloads: []Function{
			{
				Index:     0,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint8, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.MinusUint[uint8],
			},
			{
				Index:     1,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint16, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.MinusUint[uint16],
			},
			{
				Index:     2,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint32, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.MinusUint[uint32],
			},
			{
				Index:     3,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint64, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.MinusUint[uint64],
			},
			{
				Index:     4,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int8, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.MinusInt[int8],
			},
			{
				Index:     5,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int16, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.MinusInt[int16],
			},
			{
				Index:     6,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int32, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.MinusInt[int32],
			},
			{
				Index:     7,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int64, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.MinusInt[int64],
			},
			{
				Index:     8,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_float32, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.MinusFloat[float32],
			},
			{
				Index:     9,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_float64, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.MinusFloat[float64],
			},
			{
				Index:     10,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_decimal64, types.T_decimal64},
				ReturnTyp: types.T_decimal64,
				Fn:        operator.MinusDecimal64,
			},
			{
				Index:     11,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_decimal128, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.MinusDecimal128,
			},
			{
				Index:     12,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_date, types.T_interval},
				ReturnTyp: types.T_date,
				Fn:        nil,
			},
		},
	},

	MULTI: {
		Id:          MULTI,
		TypeCheckFn: GeneralBinaryOperatorTypeCheckFn1,
		Overloads: []Function{
			{
				Index:     0,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint8, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.MultUint[uint8],
			},
			{
				Index:     1,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint16, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.MultUint[uint16],
			},
			{
				Index:     2,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint32, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.MultUint[uint32],
			},
			{
				Index:     3,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint64, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.MultUint[uint64],
			},
			{
				Index:     4,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int8, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.MultInt[int8],
			},
			{
				Index:     5,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int16, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.MultInt[int16],
			},
			{
				Index:     6,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int32, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.MultInt[int32],
			},
			{
				Index:     7,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int64, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.MultInt[int64],
			},
			{
				Index:     8,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_float32, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.MultFloat[float32],
			},
			{
				Index:     9,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_float64, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.MultFloat[float64],
			},
			{
				Index:     10,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_decimal64, types.T_decimal64},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.MultDecimal64,
			},
			{
				Index:     11,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_decimal128, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.MultDecimal128,
			},
		},
	},

	DIV: {
		Id:          DIV,
		TypeCheckFn: GeneralBinaryOperatorTypeCheckFn2,
		Overloads: []Function{
			{
				Index:     0,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_float32, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.DivFloat[float32],
			},
			{
				Index:     1,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_float64, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.DivFloat[float64],
			},
			{
				Index:     2,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_decimal64, types.T_decimal64},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.DivDecimal64,
			},
			{
				Index:     3,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_decimal128, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.DivDecimal128,
			},
		},
	},

	INTEGER_DIV: {
		Id:          INTEGER_DIV,
		TypeCheckFn: GeneralBinaryOperatorTypeCheckFn2,
		Overloads: []Function{
			{
				Index:     0,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_float32, types.T_float32},
				ReturnTyp: types.T_int64,
				Fn:        operator.IntegerDivFloat[float32],
			},
			{
				Index:     1,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_float64, types.T_float64},
				ReturnTyp: types.T_int64,
				Fn:        operator.IntegerDivFloat[float64],
			},
		},
	},

	MOD: {
		Id:          MOD,
		TypeCheckFn: GeneralBinaryOperatorTypeCheckFn1,
		Overloads: []Function{
			{
				Index:     0,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint8, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.ModUint[uint8],
			},
			{
				Index:     1,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint16, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.ModUint[uint16],
			},
			{
				Index:     2,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint32, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.ModUint[uint32],
			},
			{
				Index:     3,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint64, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.ModUint[uint64],
			},
			{
				Index:     4,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int8, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.ModInt[int8],
			},
			{
				Index:     5,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int16, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.ModInt[int16],
			},
			{
				Index:     6,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int32, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.ModInt[int32],
			},
			{
				Index:     7,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int64, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.ModInt[int64],
			},
			{
				Index:     8,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_float32, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.ModFloat[float32],
			},
			{
				Index:     9,
				Flag:      plan.Function_STRICT,
				Layout:    BINARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_float64, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.ModFloat[float64],
			},
		},
	},

	UNARY_PLUS: {
		Id: UNARY_PLUS,
		Overloads: []Function{
			{
				Index:     0,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn: func(vs []*vector.Vector, proc *process.Process) (*vector.Vector, error) {
					return vs[0], nil
				},
			},
			{
				Index:     1,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn: func(vs []*vector.Vector, proc *process.Process) (*vector.Vector, error) {
					return vs[0], nil
				},
			},
			{
				Index:     2,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn: func(vs []*vector.Vector, proc *process.Process) (*vector.Vector, error) {
					return vs[0], nil
				},
			},
			{
				Index:     3,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn: func(vs []*vector.Vector, proc *process.Process) (*vector.Vector, error) {
					return vs[0], nil
				},
			},
			{
				Index:     4,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int8},
				ReturnTyp: types.T_int8,
				Fn: func(vs []*vector.Vector, proc *process.Process) (*vector.Vector, error) {
					return vs[0], nil
				},
			},
			{
				Index:     5,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int16},
				ReturnTyp: types.T_int16,
				Fn: func(vs []*vector.Vector, proc *process.Process) (*vector.Vector, error) {
					return vs[0], nil
				},
			},
			{
				Index:     6,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int32},
				ReturnTyp: types.T_int32,
				Fn: func(vs []*vector.Vector, proc *process.Process) (*vector.Vector, error) {
					return vs[0], nil
				},
			},
			{
				Index:     7,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int64},
				ReturnTyp: types.T_int64,
				Fn: func(vs []*vector.Vector, proc *process.Process) (*vector.Vector, error) {
					return vs[0], nil
				},
			},
			{
				Index:     8,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_float32},
				ReturnTyp: types.T_float32,
				Fn: func(vs []*vector.Vector, proc *process.Process) (*vector.Vector, error) {
					return vs[0], nil
				},
			},
			{
				Index:     9,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_float64},
				ReturnTyp: types.T_float64,
				Fn: func(vs []*vector.Vector, proc *process.Process) (*vector.Vector, error) {
					return vs[0], nil
				},
			},
			{
				Index:     10,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_decimal64},
				ReturnTyp: types.T_decimal64,
				Fn: func(vs []*vector.Vector, proc *process.Process) (*vector.Vector, error) {
					return vs[0], nil
				},
			},
			{
				Index:     11,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn: func(vs []*vector.Vector, proc *process.Process) (*vector.Vector, error) {
					return vs[0], nil
				},
			},
		},
	},

	UNARY_MINUS: {
		Id: UNARY_MINUS,
		Overloads: []Function{
			{
				Index:     0,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.UnaryMinus[int8],
			},
			{
				Index:     1,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.UnaryMinus[int16],
			},
			{
				Index:     2,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.UnaryMinus[int32],
			},
			{
				Index:     3,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.UnaryMinus[int64],
			},
			{
				Index:     4,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.UnaryMinus[float32],
			},
			{
				Index:     5,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.UnaryMinus[float64],
			},
			{
				Index:     6,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_decimal64},
				ReturnTyp: types.T_decimal64,
				Fn:        operator.UnaryMinusDecimal64,
			},
			{
				Index:     7,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.UnaryMinusDecimal128,
			},
		},
	},

	UNARY_TILDE: {
		Id: UNARY_TILDE,
		Overloads: []Function{
			{
				Index:     0,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int8},
				ReturnTyp: types.T_uint64,
				Fn:        operator.UnaryTilde[int8],
			},
			{
				Index:     1,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int16},
				ReturnTyp: types.T_uint64,
				Fn:        operator.UnaryTilde[int16],
			},
			{
				Index:     2,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int32},
				ReturnTyp: types.T_uint64,
				Fn:        operator.UnaryTilde[int32],
			},
			{
				Index:     3,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_int64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.UnaryTilde[int64],
			},
			{
				Index:     4,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint8},
				ReturnTyp: types.T_uint64,
				Fn:        operator.UnaryTilde[uint8],
			},
			{
				Index:     5,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint16},
				ReturnTyp: types.T_uint64,
				Fn:        operator.UnaryTilde[uint16],
			},
			{
				Index:     6,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint32},
				ReturnTyp: types.T_uint64,
				Fn:        operator.UnaryTilde[uint32],
			},
			{
				Index:     7,
				Flag:      plan.Function_STRICT,
				Layout:    UNARY_ARITHMETIC_OPERATOR,
				Args:      []types.T{types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.UnaryTilde[uint64],
			},
		},
	},

	// others
	CAST: {
		Id: CAST,
		TypeCheckFn: func(overloads []Function, inputs []types.T) (overloadIndex int32, _ []types.T) {
			// cast-operator should check param types strictly
			if len(inputs) == 2 {
				for i, o := range overloads {
					if o.Args[0] == inputs[0] && o.Args[1] == inputs[1] {
						return int32(i), nil
					}
				}
			}
			return wrongFunctionParameters, nil
		},
		Overloads: []Function{
			{
				Index:     0,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int8, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.Cast,
			},
			{
				Index:     1,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int16, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.Cast,
			},
			{
				Index:     2,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int32, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.Cast,
			},
			{
				Index:     3,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int64, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.Cast,
			},
			{
				Index:     4,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint8, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.Cast,
			},
			{
				Index:     5,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint16, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.Cast,
			},
			{
				Index:     6,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint32, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.Cast,
			},
			{
				Index:     7,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint64, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.Cast,
			},
			{
				Index:     8,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float32, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.Cast,
			},
			{
				Index:     9,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float64, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.Cast,
			},
			{
				Index:     10,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_date, types.T_date},
				ReturnTyp: types.T_date,
				Fn:        operator.Cast,
			},
			{
				Index:     11,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_datetime, types.T_datetime},
				ReturnTyp: types.T_datetime,
				Fn:        operator.Cast,
			},
			{
				Index:     12,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_timestamp, types.T_timestamp},
				ReturnTyp: types.T_timestamp,
				Fn:        operator.Cast,
			},
			{
				Index:     13,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int16, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.Cast,
			},
			{
				Index:     14,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int32, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.Cast,
			},
			{
				Index:     15,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int64, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.Cast,
			},
			{
				Index:     16,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint8, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.Cast,
			},
			{
				Index:     17,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint16, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.Cast,
			},
			{
				Index:     18,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint32, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.Cast,
			},
			{
				Index:     19,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint64, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.Cast,
			},
			{
				Index:     20,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float32, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.Cast,
			},
			{
				Index:     21,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float64, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.Cast,
			},
			{
				Index:     22,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int8, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.Cast,
			},
			{
				Index:     23,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int32, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.Cast,
			},
			{
				Index:     24,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int64, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.Cast,
			},
			{
				Index:     25,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint8, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.Cast,
			},
			{
				Index:     26,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint16, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.Cast,
			},
			{
				Index:     27,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint32, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.Cast,
			},
			{
				Index:     28,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint64, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.Cast,
			},
			{
				Index:     29,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float32, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.Cast,
			},
			{
				Index:     30,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float64, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.Cast,
			},
			{
				Index:     31,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int8, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.Cast,
			},
			{
				Index:     32,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int16, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.Cast,
			},
			{
				Index:     33,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int64, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.Cast,
			},
			{
				Index:     34,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint8, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.Cast,
			},
			{
				Index:     35,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint16, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.Cast,
			},
			{
				Index:     36,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint32, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.Cast,
			},
			{
				Index:     37,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint64, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.Cast,
			},
			{
				Index:     38,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float32, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.Cast,
			},
			{
				Index:     39,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float64, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.Cast,
			},
			{
				Index:     40,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int8, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.Cast,
			},
			{
				Index:     41,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int16, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.Cast,
			},
			{
				Index:     42,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int32, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.Cast,
			},
			{
				Index:     43,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint8, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.Cast,
			},
			{
				Index:     44,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint16, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.Cast,
			},
			{
				Index:     45,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint32, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.Cast,
			},
			{
				Index:     46,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint64, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.Cast,
			},
			{
				Index:     47,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float32, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.Cast,
			},
			{
				Index:     48,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float64, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.Cast,
			},
			{
				Index:     49,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int8, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.Cast,
			},
			{
				Index:     50,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int16, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.Cast,
			},
			{
				Index:     51,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int32, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.Cast,
			},
			{
				Index:     52,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int64, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.Cast,
			},
			{
				Index:     53,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint16, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.Cast,
			},
			{
				Index:     54,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint32, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.Cast,
			},
			{
				Index:     55,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint64, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.Cast,
			},
			{
				Index:     56,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float32, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.Cast,
			},
			{
				Index:     57,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float64, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.Cast,
			},
			{
				Index:     58,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int8, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.Cast,
			},
			{
				Index:     59,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int16, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.Cast,
			},
			{
				Index:     60,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int32, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.Cast,
			},
			{
				Index:     61,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int64, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.Cast,
			},
			{
				Index:     62,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint8, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.Cast,
			},
			{
				Index:     63,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint32, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.Cast,
			},
			{
				Index:     64,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint64, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.Cast,
			},
			{
				Index:     65,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float32, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.Cast,
			},
			{
				Index:     66,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float64, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.Cast,
			},
			{
				Index:     67,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int8, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.Cast,
			},
			{
				Index:     68,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int16, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.Cast,
			},
			{
				Index:     69,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int32, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.Cast,
			},
			{
				Index:     70,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int64, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.Cast,
			},
			{
				Index:     71,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint8, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.Cast,
			},
			{
				Index:     72,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint16, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.Cast,
			},
			{
				Index:     73,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint64, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.Cast,
			},
			{
				Index:     74,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float32, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.Cast,
			},
			{
				Index:     75,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float64, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.Cast,
			},
			{
				Index:     76,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int8, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.Cast,
			},
			{
				Index:     77,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int16, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.Cast,
			},
			{
				Index:     78,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int32, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.Cast,
			},
			{
				Index:     79,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int64, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.Cast,
			},
			{
				Index:     80,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint8, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.Cast,
			},
			{
				Index:     81,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint16, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.Cast,
			},
			{
				Index:     82,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint32, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.Cast,
			},
			{
				Index:     83,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float32, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.Cast,
			},
			{
				Index:     84,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float64, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.Cast,
			},
			{
				Index:     85,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int8, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.Cast,
			},
			{
				Index:     86,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int16, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.Cast,
			},
			{
				Index:     87,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int32, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.Cast,
			},
			{
				Index:     88,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int64, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.Cast,
			},
			{
				Index:     89,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint8, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.Cast,
			},
			{
				Index:     90,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint16, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.Cast,
			},
			{
				Index:     91,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint32, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.Cast,
			},
			{
				Index:     92,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint64, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.Cast,
			},
			{
				Index:     93,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float64, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.Cast,
			},
			{
				Index:     94,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int8, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.Cast,
			},
			{
				Index:     95,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int16, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.Cast,
			},
			{
				Index:     96,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int32, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.Cast,
			},
			{
				Index:     97,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int64, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.Cast,
			},
			{
				Index:     98,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint8, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.Cast,
			},
			{
				Index:     99,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint16, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.Cast,
			},
			{
				Index:     100,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint32, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.Cast,
			},
			{
				Index:     101,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint64, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.Cast,
			},
			{
				Index:     102,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float32, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.Cast,
			},
			{
				Index:     103,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.Cast,
			},
			{
				Index:     104,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.Cast,
			},
			{
				Index:     105,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.Cast,
			},
			{
				Index:     106,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.Cast,
			},
			{
				Index:     107,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.Cast,
			},
			{
				Index:     108,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.Cast,
			},
			{
				Index:     109,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.Cast,
			},
			{
				Index:     110,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.Cast,
			},
			{
				Index:     111,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.Cast,
			},
			{
				Index:     112,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.Cast,
			},
			{
				Index:     113,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.Cast,
			},
			{
				Index:     114,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.Cast,
			},
			{
				Index:     115,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.Cast,
			},
			{
				Index:     116,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.Cast,
			},
			{
				Index:     117,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.Cast,
			},
			{
				Index:     118,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.Cast,
			},
			{
				Index:     119,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.Cast,
			},
			{
				Index:     120,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.Cast,
			},
			{
				Index:     121,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.Cast,
			},
			{
				Index:     122,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.Cast,
			},
			{
				Index:     123,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int8, types.T_char},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     124,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int8, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     125,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int16, types.T_char},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     126,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int16, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     127,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int32, types.T_char},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     128,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int32, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     129,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int64, types.T_char},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     130,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int64, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     131,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint8, types.T_char},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     132,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint8, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     133,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint16, types.T_char},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     134,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint16, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     135,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint32, types.T_char},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     136,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint32, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     137,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint64, types.T_char},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     138,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint64, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     139,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float32, types.T_char},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     140,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float32, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     141,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float64, types.T_char},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     142,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float64, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     143,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_char},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     144,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     145,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     146,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_char},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     147,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int8, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.Cast,
			},
			{
				Index:     148,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int16, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.Cast,
			},
			{
				Index:     149,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int32, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.Cast,
			},
			{
				Index:     150,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int64, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.Cast,
			},
			{
				Index:     151,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_date},
				ReturnTyp: types.T_date,
				Fn:        operator.Cast,
			},
			{
				Index:     152,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_datetime},
				ReturnTyp: types.T_datetime,
				Fn:        operator.Cast,
			},
			{
				Index:     153,
				Volatile:  true,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_timestamp},
				ReturnTyp: types.T_timestamp,
				Fn:        operator.Cast,
			},
			{
				Index:     154,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal64, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.Cast,
			},
			{
				Index:     155,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal64, types.T_decimal64},
				ReturnTyp: types.T_decimal64,
				Fn:        operator.Cast,
			},
			{
				Index:     156,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal128, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.Cast,
			},
			{
				Index:     157,
				Volatile:  true,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_timestamp, types.T_datetime},
				ReturnTyp: types.T_datetime,
				Fn:        operator.Cast,
			},
			{
				Index:     158,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_interval},
				ReturnTyp: types.T_interval,
				Fn:        nil,
			},
			{
				Index:     159,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint8, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.Cast,
			},
			{
				Index:     160,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint16, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.Cast,
			},
			{
				Index:     161,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint32, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.Cast,
			},
			{
				Index:     162,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint64, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.Cast,
			},
			{
				Index:     163,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_decimal64},
				ReturnTyp: types.T_decimal64,
				Fn:        operator.Cast,
			},
			{
				Index:     164,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_decimal64},
				ReturnTyp: types.T_decimal64,
				Fn:        operator.Cast,
			},
			{
				Index:     165,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.Cast,
			},
			{
				Index:     166,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_decimal128},
				ReturnTyp: types.T_decimal64,
				Fn:        operator.Cast,
			},
			{
				Index:     167,
				Volatile:  true,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_timestamp, types.T_varchar},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     168,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_date, types.T_char},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     169,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_date, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     170,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_datetime, types.T_char},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     171,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_datetime, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     172,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_bool, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     173,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_bool, types.T_char},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     174,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_date, types.T_datetime},
				ReturnTyp: types.T_datetime,
				Fn:        operator.Cast,
			},
			{
				Index:     175,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_datetime, types.T_date},
				ReturnTyp: types.T_date,
				Fn:        operator.Cast,
			},
			{
				Index:     176,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float32, types.T_decimal64},
				ReturnTyp: types.T_decimal64,
				Fn:        operator.Cast,
			},
			{
				Index:     177,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float64, types.T_decimal64},
				ReturnTyp: types.T_decimal64,
				Fn:        operator.Cast,
			},
			{
				Index:     178,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float64, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.Cast,
			},
			{
				Index:     179,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float32, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.Cast,
			},
			{
				Index:     180,
				Volatile:  true,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_datetime, types.T_timestamp},
				ReturnTyp: types.T_timestamp,
				Fn:        operator.Cast,
			},
			{
				Index:     181,
				Volatile:  true,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_date, types.T_timestamp},
				ReturnTyp: types.T_timestamp,
				Fn:        operator.Cast,
			},
			{
				Index:     182,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal64, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     183,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal128, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     184,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal64, types.T_char},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     185,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal128, types.T_char},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     186,
				Volatile:  true,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_timestamp, types.T_date},
				ReturnTyp: types.T_date,
				Fn:        operator.Cast,
			},
			{
				Index:     187,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int8, types.T_timestamp},
				ReturnTyp: types.T_timestamp,
				Fn:        operator.Cast,
			},
			{
				Index:     188,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int16, types.T_timestamp},
				ReturnTyp: types.T_timestamp,
				Fn:        operator.Cast,
			},
			{
				Index:     189,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int32, types.T_timestamp},
				ReturnTyp: types.T_timestamp,
				Fn:        operator.Cast,
			},
			{
				Index:     190,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int64, types.T_timestamp},
				ReturnTyp: types.T_timestamp,
				Fn:        operator.Cast,
			},
			{
				Index:     191,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint8, types.T_timestamp},
				ReturnTyp: types.T_timestamp,
				Fn:        operator.Cast,
			},
			{
				Index:     192,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint16, types.T_timestamp},
				ReturnTyp: types.T_timestamp,
				Fn:        operator.Cast,
			},
			{
				Index:     193,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint32, types.T_timestamp},
				ReturnTyp: types.T_timestamp,
				Fn:        operator.Cast,
			},
			{
				Index:     194,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint64, types.T_timestamp},
				ReturnTyp: types.T_timestamp,
				Fn:        operator.Cast,
			},
			{
				Index:     195,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal64, types.T_timestamp},
				ReturnTyp: types.T_timestamp,
				Fn:        operator.Cast,
			},
			{
				Index:     196,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal128, types.T_timestamp},
				ReturnTyp: types.T_timestamp,
				Fn:        operator.Cast,
			},
			{
				Index:     197,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal64, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.Cast,
			},
			{
				Index:     198,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal128, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.Cast,
			},
			{
				Index:     199,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal64, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.Cast,
			},
			{
				Index:     200,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal128, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.Cast,
			},
			{
				Index:     201,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_date},
				ReturnTyp: types.T_date,
				Fn:        operator.Cast,
			},
			{
				Index:     202,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_datetime},
				ReturnTyp: types.T_datetime,
				Fn:        operator.Cast,
			},
			{
				Index:     203,
				Volatile:  true,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_timestamp},
				ReturnTyp: types.T_timestamp,
				Fn:        operator.Cast,
			},
			{
				Index:     204,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int8, types.T_decimal64},
				ReturnTyp: types.T_decimal64,
				Fn:        operator.Cast,
			},
			{
				Index:     205,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int16, types.T_decimal64},
				ReturnTyp: types.T_decimal64,
				Fn:        operator.Cast,
			},
			{
				Index:     206,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int32, types.T_decimal64},
				ReturnTyp: types.T_decimal64,
				Fn:        operator.Cast,
			},
			{
				Index:     207,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int64, types.T_decimal64},
				ReturnTyp: types.T_decimal64,
				Fn:        operator.Cast,
			},
			{
				Index:     208,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int8, types.T_bool},
				ReturnTyp: types.T_bool,
				Fn:        operator.Cast,
			},
			{
				Index:     209,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int16, types.T_bool},
				ReturnTyp: types.T_bool,
				Fn:        operator.Cast,
			},
			{
				Index:     210,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int32, types.T_bool},
				ReturnTyp: types.T_bool,
				Fn:        operator.Cast,
			},
			{
				Index:     211,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int64, types.T_bool},
				ReturnTyp: types.T_bool,
				Fn:        operator.Cast,
			},
			{
				Index:     212,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint8, types.T_bool},
				ReturnTyp: types.T_bool,
				Fn:        operator.Cast,
			},
			{
				Index:     213,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint16, types.T_bool},
				ReturnTyp: types.T_bool,
				Fn:        operator.Cast,
			},
			{
				Index:     214,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint32, types.T_bool},
				ReturnTyp: types.T_bool,
				Fn:        operator.Cast,
			},
			{
				Index:     215,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint64, types.T_bool},
				ReturnTyp: types.T_bool,
				Fn:        operator.Cast,
			},
			{
				Index:     216,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float32, types.T_bool},
				ReturnTyp: types.T_bool,
				Fn:        operator.Cast,
			},
			{
				Index:     217,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float64, types.T_bool},
				ReturnTyp: types.T_bool,
				Fn:        operator.Cast,
			},
			{
				Index:     218,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_bool},
				ReturnTyp: types.T_bool,
				Fn:        operator.Cast,
			},
			{
				Index:     219,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_bool},
				ReturnTyp: types.T_bool,
				Fn:        operator.Cast,
			},
			{
				Index:     220,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal64, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.Cast,
			},
			{
				Index:     221,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal128, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.Cast,
			},
			{
				Index:     222,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal64, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.Cast,
			},
			{
				Index:     223,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal128, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.Cast,
			},
			{
				Index:     224,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal64, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.Cast,
			},
			{
				Index:     225,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal128, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.Cast,
			},
			{
				Index:     226,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal128, types.T_decimal64},
				ReturnTyp: types.T_decimal64,
				Fn:        operator.Cast,
			},
			{
				Index:     227,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.Cast,
			},

			{
				Index:     228,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.Cast,
			},
			{
				Index:     229,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.Cast,
			},
			{
				Index:     230,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.Cast,
			},
			{
				Index:     231,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.Cast,
			},
			{
				Index:     232,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.Cast,
			},
			{
				Index:     233,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.Cast,
			},
			{
				Index:     234,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.Cast,
			},
			{
				Index:     235,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_float32},
				ReturnTyp: types.T_float32,
				Fn:        operator.Cast,
			},
			{
				Index:     236,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_float64},
				ReturnTyp: types.T_float64,
				Fn:        operator.Cast,
			},
			{
				Index:     237,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int8, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     238,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int16, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     239,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int32, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     240,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_int64, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     241,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint8, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     242,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint16, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     243,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint32, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     244,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_uint64, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     245,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float32, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     246,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_float64, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     247,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     248,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_varchar},
				ReturnTyp: types.T_varchar,
				Fn:        operator.Cast,
			},
			{
				Index:     249,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_char, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     250,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_char},
				ReturnTyp: types.T_char,
				Fn:        operator.Cast,
			},
			{
				Index:     251,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     252,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_date},
				ReturnTyp: types.T_date,
				Fn:        operator.Cast,
			},
			{
				Index:     253,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_datetime},
				ReturnTyp: types.T_datetime,
				Fn:        operator.Cast,
			},
			{
				Index:     254,
				Volatile:  true,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_timestamp},
				ReturnTyp: types.T_timestamp,
				Fn:        operator.Cast,
			},
			{
				Index:     255,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_decimal64},
				ReturnTyp: types.T_decimal64,
				Fn:        operator.Cast,
			},
			{
				Index:     256,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_decimal128},
				ReturnTyp: types.T_decimal128,
				Fn:        operator.Cast,
			},
			{
				Index:     257,
				Volatile:  true,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_timestamp, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     258,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_date, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     259,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_datetime, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     260,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_bool, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     261,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal64, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     262,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal128, types.T_blob},
				ReturnTyp: types.T_blob,
				Fn:        operator.Cast,
			},
			{
				Index:     263,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_blob, types.T_bool},
				ReturnTyp: types.T_bool,
				Fn:        operator.Cast,
			},
			{
				Index:     264,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_varchar, types.T_json},
				ReturnTyp: types.T_json,
				Fn:        operator.Cast,
			},
			{
				Index:     265,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_decimal128, types.T_int32},
				ReturnTyp: types.T_json,
				Fn:        operator.Cast,
			},
			{
				Index:     266,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_bool, types.T_int8},
				ReturnTyp: types.T_int8,
				Fn:        operator.Cast,
			},
			{
				Index:     267,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_bool, types.T_int16},
				ReturnTyp: types.T_int16,
				Fn:        operator.Cast,
			},
			{
				Index:     268,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_bool, types.T_int32},
				ReturnTyp: types.T_int32,
				Fn:        operator.Cast,
			},
			{
				Index:     269,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_bool, types.T_int64},
				ReturnTyp: types.T_int64,
				Fn:        operator.Cast,
			},
			{
				Index:     270,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_bool, types.T_uint8},
				ReturnTyp: types.T_uint8,
				Fn:        operator.Cast,
			},
			{
				Index:     271,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_bool, types.T_uint16},
				ReturnTyp: types.T_uint16,
				Fn:        operator.Cast,
			},
			{
				Index:     272,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_bool, types.T_uint32},
				ReturnTyp: types.T_uint32,
				Fn:        operator.Cast,
			},
			{
				Index:     273,
				Flag:      plan.Function_STRICT,
				Layout:    CAST_EXPRESSION,
				Args:      []types.T{types.T_bool, types.T_uint64},
				ReturnTyp: types.T_uint64,
				Fn:        operator.Cast,
			},
		},
	},

	COALESCE: {
		Id: COALESCE,
		TypeCheckFn: func(overloads []Function, inputs []types.T) (overloadIndex int32, ts []types.T) {
			l := len(inputs)
			if l == 0 {
				return wrongFunctionParameters, nil
			}

			for i, o := range overloads {
				if operator.CoalesceTypeCheckFn(inputs, nil, o.ReturnTyp) {
					return int32(i), nil
				}
			}

			minCost, minIndex := math.MaxInt32, -1
			convertTypes := make([]types.T, l)
			targetTypes := make([]types.T, l)

			for i, o := range overloads {
				for j := 0; j < l; j++ {
					targetTypes[j] = o.ReturnTyp
				}
				if code, c := tryToMatch(inputs, targetTypes); code == matchedByConvert {
					if c < minCost {
						minCost = c
						copy(convertTypes, targetTypes)
						minIndex = i
					}
				}
			}
			if minIndex != -1 {
				return int32(minIndex), convertTypes
			}
			return wrongFunctionParameters, nil
		},
		Overloads: []Function{
			{
				Index:     0,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_varchar,
				Fn:        operator.CoalesceVarchar,
			},
			{
				Index:     1,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_char,
				Fn:        operator.CoalesceChar,
			},
			{
				Index:     2,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_int8,
				Fn:        operator.CoalesceInt8,
			},
			{
				Index:     3,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_int16,
				Fn:        operator.CoalesceInt16,
			},
			{
				Index:     4,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_int32,
				Fn:        operator.CoalesceInt32,
			},
			{
				Index:     5,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_int64,
				Fn:        operator.CoalesceInt64,
			},
			{
				Index:     6,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_uint8,
				Fn:        operator.CoalesceUint8,
			},
			{
				Index:     7,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_uint16,
				Fn:        operator.CoalesceUint16,
			},
			{
				Index:     8,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_uint32,
				Fn:        operator.CoalesceUint32,
			},
			{
				Index:     9,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_uint64,
				Fn:        operator.CoalesceUint64,
			},
			{
				Index:     10,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_float32,
				Fn:        operator.CoalesceFloat32,
			},
			{
				Index:     11,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_float64,
				Fn:        operator.CoalesceFloat64,
			},
			{
				Index:     12,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_bool,
				Fn:        operator.CoalesceBool,
			},
			{
				Index:     13,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_date,
				Fn:        operator.CoalesceDate,
			},
			{
				Index:     14,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_datetime,
				Fn:        operator.CoalesceDateTime,
			},
			{
				Index:     15,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_decimal64,
				Fn:        operator.CoalesceDecimal64,
			},
			{
				Index:     16,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_decimal128,
				Fn:        operator.CoalesceDecimal128,
			},
			{
				Index:     17,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_timestamp,
				Fn:        operator.CoalesceTimestamp,
			},
		},
	},

	CASE: {
		Id: CASE,
		TypeCheckFn: func(overloads []Function, inputs []types.T) (overloadIndex int32, ts []types.T) {
			for i, o := range overloads {
				if operator.CwTypeCheckFn(inputs, nil, o.ReturnTyp) {
					return int32(i), nil
				}
			}
			l := len(inputs)
			minCost, minIndex := math.MaxInt32, -1
			convertTypes := make([]types.T, l)
			targetTypes := make([]types.T, l)
			for i, o := range overloads {
				if l >= 2 {
					flag := true
					for j := 0; j < l-1; j += 2 {
						if inputs[j] != types.T_bool && !inputs[0].ToType().IsIntOrUint() {
							flag = false
							break
						}
						targetTypes[j] = types.T_bool
					}
					if l%2 == 1 {
						targetTypes[l-1] = o.ReturnTyp
					}
					for j := 1; j < l; j += 2 {
						targetTypes[j] = o.ReturnTyp
					}
					if flag {
						if code, c := tryToMatch(inputs, targetTypes); code == matchedByConvert {
							if c < minCost {
								minCost = c
								copy(convertTypes, targetTypes)
								minIndex = i
							}
						}
					}
				}
			}
			if minIndex != -1 {
				return int32(minIndex), convertTypes
			}
			return wrongFunctionParameters, nil
		},
		Overloads: []Function{
			{
				Index:     0,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_int8,
				Fn:        operator.CaseWhenInt8,
			},
			{
				Index:     1,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_int16,
				Fn:        operator.CaseWhenInt16,
			},
			{
				Index:     2,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_int32,
				Fn:        operator.CaseWhenInt32,
			},
			{
				Index:     3,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_int64,
				Fn:        operator.CaseWhenInt64,
			},
			{
				Index:     4,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_uint8,
				Fn:        operator.CaseWhenUint8,
			},
			{
				Index:     5,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_uint16,
				Fn:        operator.CaseWhenUint16,
			},
			{
				Index:     6,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_uint32,
				Fn:        operator.CaseWhenUint32,
			},
			{
				Index:     7,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_uint64,
				Fn:        operator.CaseWhenUint64,
			},
			{
				Index:     8,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_float32,
				Fn:        operator.CaseWhenFloat32,
			},
			{
				Index:     9,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_float64,
				Fn:        operator.CaseWhenFloat64,
			},
			{
				Index:     10,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_bool,
				Fn:        operator.CaseWhenBool,
			},
			{
				Index:     11,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_date,
				Fn:        operator.CaseWhenDate,
			},
			{
				Index:     12,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_datetime,
				Fn:        operator.CaseWhenDateTime,
			},
			{
				Index:     13,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_varchar,
				Fn:        operator.CaseWhenVarchar,
			},
			{
				Index:     14,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_char,
				Fn:        operator.CaseWhenChar,
			},
			{
				Index:     15,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_decimal64,
				Fn:        operator.CaseWhenDecimal64,
			},
			{
				Index:     16,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_decimal128,
				Fn:        operator.CaseWhenDecimal128,
			},
			{
				Index:     17,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_timestamp,
				Fn:        operator.CaseWhenTimestamp,
			},
			{
				Index:     18,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_blob,
				Fn:        operator.CaseWhenText,
			},
		},
	},

	IFF: {
		Id: IFF,
		TypeCheckFn: func(overloads []Function, inputs []types.T) (overloadIndex int32, ts []types.T) {
			for i, o := range overloads {
				if operator.IfTypeCheckFn(inputs, nil, o.ReturnTyp) {
					return int32(i), nil
				}
			}
			minCost, minIndex := math.MaxInt32, -1
			convertTypes := make([]types.T, 3)
			targetTypes := make([]types.T, 3)
			for i, o := range overloads {
				if len(inputs) == 3 {
					if inputs[0] != types.T_bool && !inputs[0].ToType().IsIntOrUint() {
						continue
					}

					targetTypes[0] = types.T_bool
					targetTypes[1], targetTypes[2] = o.ReturnTyp, o.ReturnTyp
					if code, c := tryToMatch(inputs, targetTypes); code == matchedByConvert {
						if c < minCost {
							minCost = c
							copy(convertTypes, targetTypes)
							minIndex = i
						}
					}
				}
			}
			if minIndex != -1 {
				return int32(minIndex), convertTypes
			}
			return wrongFunctionParameters, nil
		},
		Overloads: []Function{
			{
				Index:     0,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_int8,
				Fn:        operator.IfInt8,
			},
			{
				Index:     1,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_int16,
				Fn:        operator.IfInt16,
			},
			{
				Index:     2,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_int32,
				Fn:        operator.IfInt32,
			},
			{
				Index:     3,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_int64,
				Fn:        operator.IfInt64,
			},
			{
				Index:     4,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_uint8,
				Fn:        operator.IfUint8,
			},
			{
				Index:     5,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_uint16,
				Fn:        operator.IfUint16,
			},
			{
				Index:     6,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_uint32,
				Fn:        operator.IfUint32,
			},
			{
				Index:     7,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_uint64,
				Fn:        operator.IfUint64,
			},
			{
				Index:     8,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_float32,
				Fn:        operator.IfFloat32,
			},
			{
				Index:     9,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_float64,
				Fn:        operator.IfFloat64,
			},
			{
				Index:     10,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_bool,
				Fn:        operator.IfBool,
			},
			{
				Index:     11,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_date,
				Fn:        operator.IfDate,
			},
			{
				Index:     12,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_datetime,
				Fn:        operator.IfDateTime,
			},
			{
				Index:     13,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_varchar,
				Fn:        operator.IfVarchar,
			},
			{
				Index:     14,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_char,
				Fn:        operator.IfChar,
			},
			{
				Index:     15,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_decimal64,
				Fn:        operator.IfDecimal64,
			},
			{
				Index:     16,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_decimal128,
				Fn:        operator.IfDecimal128,
			},
			{
				Index:     17,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    CASE_WHEN_EXPRESSION,
				ReturnTyp: types.T_timestamp,
				Fn:        operator.IfTimestamp,
			},
			{
				Index:     18,
				Volatile:  true,
				Flag:      plan.Function_NONE,
				Layout:    STANDARD_FUNCTION,
				ReturnTyp: types.T_blob,
				Fn:        operator.IfText,
			},
		},
	},
}
