// Copyright 2022 Matrix Origin
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//	http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package metric

import (
	"context"
	"sync"
	"testing"
	"time"

	pb "github.com/matrixorigin/matrixone/pkg/pb/metric"
	prom "github.com/prometheus/client_golang/prometheus"
	"github.com/stretchr/testify/assert"
)

type dummyCollect struct {
	dummySwitch
	sync.Mutex
	mfs [][]*pb.MetricFamily
}

func (e *dummyCollect) SendMetrics(ctx context.Context, mfs []*pb.MetricFamily) error {
	e.Lock()
	defer e.Unlock()
	e.mfs = append(e.mfs, mfs)
	return nil
}

func (e *dummyCollect) sendCnt() int {
	e.Lock()
	defer e.Unlock()
	return len(e.mfs)
}

func TestExporterCommonInfo(t *testing.T) {
	exp := newMetricExporter(nil, nil, "node_uuid_42", "monolithic").(*metricExporter)
	mfs := []*pb.MetricFamily{
		{Metric: []*pb.Metric{{}}},
		{Metric: []*pb.Metric{{Label: []*pb.LabelPair{{Name: "color", Value: "blue"}}}}},
		{Metric: []*pb.Metric{{Label: []*pb.LabelPair{{Name: "color", Value: "red"}, {Name: "zaxis", Value: "10"}}}}},
		{Metric: []*pb.Metric{
			{Label: []*pb.LabelPair{
				{Name: "color", Value: "red"},
				{Name: "zaxis", Value: "10"},
				{Name: "env", Value: "test"},
			}},
			{Label: []*pb.LabelPair{
				{Name: "color", Value: "gray"},
				{Name: "zaxis", Value: "20"},
				{Name: "env", Value: "dev"},
			}},
		}},
	}
	exp.addCommonInfo(mfs)
	if mfs[0].Metric[0].Collecttime == 0 {
		t.Error("addCommonInfo adds no collecttime")
	}
	time := mfs[0].Metric[0].GetCollecttime()

	names := []string{"color", "zaxis", "env"}
	lblCnt := 0
	for i, mf := range mfs {
		assert.Equal(t, mf.GetNode(), "node_uuid_42")
		assert.Equal(t, mf.GetRole(), "monolithic")
		name := names[:lblCnt]
		for _, m := range mf.Metric {
			if m.GetCollecttime() != time {
				t.Errorf("addCommonInfo adds different collecttime, group %d", i)
			}
			if len(m.Label) != lblCnt {
				t.Errorf("addCommonInfo add wrong labels, group %d", i)
			}
			for i := range m.Label {
				if m.Label[i].Name != name[i] {
					t.Errorf("addCommonInfo add wrong labels, want %s@%d, got %s", name[i], i, m.Label[i])
				}
			}
		}
		lblCnt += 1
	}
}

func TestExporter(t *testing.T) {
	dumCollect := &dummyCollect{}
	dumClock := makeDummyClock(1)
	var exp *metricExporter

	withModifiedConfig(func() {
		defer setGatherInterval(setGatherInterval(20 * time.Millisecond))
		defer setRawHistBufLimit(setRawHistBufLimit(5))
		defer setExportToProm(setExportToProm(false))
		reg := prom.NewRegistry()
		iexp := newMetricExporter(reg, dumCollect, "node_uuid", "monolithic")
		exp = iexp.(*metricExporter)
		exp.Start(context.TODO())
		defer exp.Stop(false)
		exp.now = dumClock
		c := prom.NewCounter(prom.CounterOpts{Subsystem: "test", Name: "test_counter"})
		reg.MustRegister(c)
		g := prom.NewGauge(prom.GaugeOpts{Subsystem: "test", Name: "test_gauge"})
		reg.MustRegister(g)
		h := NewRawHist(prom.HistogramOpts{Subsystem: "test", Name: "test_hist"})
		h.exporter = &iexp
		h.now = dumClock
		reg.MustRegister(h)

		wg := new(sync.WaitGroup)
		wg.Add(1)
		go func() {
			// ~45ms, 2 rawHist full export
			for i := 0; i < 14; i++ {
				c.Add(1)
				g.Add(2.0)
				h.Observe(float64(i))
			}
			// wait 2 Gather()
			time.Sleep(50 * time.Millisecond)
			wg.Done()
		}()
		wg.Wait()
	})
	sendCnt := dumCollect.sendCnt()
	time.Sleep(25 * time.Millisecond)
	if dumCollect.sendCnt() != sendCnt {
		t.Error("collector receive metrics after stopping")
	}
	if sendCnt != 4 {
		// test involving timer is not stable, if not matched just return. Fix it later
		// t.Errorf("collector receive %d batch metrics, want 4", sendCnt)
		t.Logf("[Metric TODO]: collector receive %d batch metrics, want 4", sendCnt)
		return
	}

	// 14 Observe + 4 addCommonInfo
	if dumClock()-1 != 14+4 {
		t.Errorf("disorder time")
	}

	dumCollect.Lock()

	counterCnt, gaugeCnt, sampleCnt := 0, 0, 0
	for _, mfs := range dumCollect.mfs {
		for _, mf := range mfs {
			switch mf.GetType() {
			case pb.MetricType_COUNTER:
				counterCnt += len(mf.Metric)
			case pb.MetricType_GAUGE:
				gaugeCnt += len(mf.Metric)
			case pb.MetricType_RAWHIST:
				for _, m := range mf.Metric {
					sampleCnt += len(m.RawHist.Samples)
				}
			}
		}
	}
	if counterCnt != 2 || gaugeCnt != 2 {
		t.Error("mismatched counters and gauges")
	}

	if sampleCnt != 14 {
		t.Errorf("mismatched hist samples, want 14, got %d", sampleCnt)
	}
	dumCollect.Unlock()

	func() {
		exp.Lock()
		defer exp.Unlock()
		if cap(exp.histFamilies) != 1 {
			t.Error("exporter buffer capacity should be 1")
		}
	}()
}
