// Copyright 2022 Matrix Origin
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package indexwrapper

import (
	"github.com/RoaringBitmap/roaring"
	"github.com/matrixorigin/matrixone/pkg/container/types"
	"github.com/matrixorigin/matrixone/pkg/vm/engine/tae/catalog"
	"github.com/matrixorigin/matrixone/pkg/vm/engine/tae/containers"
	"github.com/matrixorigin/matrixone/pkg/vm/engine/tae/iface/data"
	"github.com/matrixorigin/matrixone/pkg/vm/engine/tae/index"
)

type immutableIndex struct {
	zmReader *ZMReader
	bfReader *BFReader
}

func NewImmutableIndex() *immutableIndex {
	return new(immutableIndex)
}

func (index *immutableIndex) IsKeyDeleted(any, types.TS) (bool, bool)        { panic("not supported") }
func (index *immutableIndex) GetActiveRow(any) (uint32, error)               { panic("not supported") }
func (index *immutableIndex) Delete(any, types.TS) error                     { panic("not supported") }
func (index *immutableIndex) RevertUpsert(containers.Vector, types.TS) error { panic("not supported") }
func (index *immutableIndex) BatchUpsert(*index.KeysCtx, int, types.TS) (*index.BatchResp, error) {
	panic("not supported")
}

func (index *immutableIndex) Dedup(key any) (err error) {
	exist := index.zmReader.Contains(key)
	// 2. if not in [min, max], key is definitely not found
	if !exist {
		return
	}
	exist, err = index.bfReader.MayContainsKey(key)
	// 3. check bloomfilter has some error. return err
	if err != nil {
		err = TranslateError(err)
		return
	}
	if exist {
		err = data.ErrPossibleDuplicate
	}
	return
}

func (index *immutableIndex) String() string {
	panic("implement me")
}
func (index *immutableIndex) GetMaxDeleteTS() types.TS                    { panic("not supported") }
func (index *immutableIndex) HasDeleteFrom(key any, fromTs types.TS) bool { panic("not supported") }

func (index *immutableIndex) BatchDedup(keys containers.Vector,
	rowmask *roaring.Bitmap) (keyselects *roaring.Bitmap, err error) {
	keyselects, exist := index.zmReader.ContainsAny(keys)
	// 1. all keys are not in [min, max]. definitely not
	if !exist {
		return
	}
	exist, keyselects, err = index.bfReader.MayContainsAnyKeys(keys, keyselects)
	// 3. check bloomfilter has some unknown error. return err
	if err != nil {
		err = TranslateError(err)
		return
	}
	// 4. all keys were checked. definitely not
	if !exist {
		return
	}
	err = data.ErrPossibleDuplicate
	return
}

func (index *immutableIndex) Close() (err error) {
	// TODO
	return
}

func (index *immutableIndex) Destroy() (err error) {
	if index.zmReader != nil {
		if err = index.zmReader.Destroy(); err != nil {
			return
		}
	}
	if index.bfReader != nil {
		err = index.bfReader.Destroy()
	}
	return
}

func (index *immutableIndex) ReadFrom(blk data.Block) (err error) {
	entry := blk.GetMeta().(*catalog.BlockEntry)
	file := blk.GetBlockFile()
	idxMeta, err := file.LoadIndexMeta()
	if err != nil {
		return
	}
	metas := idxMeta.(*IndicesMeta)
	colDef := entry.GetSchema().SortKey.Defs[0]
	colFile, err := file.OpenColumn(colDef.Idx)
	if err != nil {
		return
	}
	defer colFile.Close()
	for _, meta := range metas.Metas {
		idxFile, err := colFile.OpenIndexFile(int(meta.InternalIdx))
		if err != nil {
			return err
		}
		id := entry.AsCommonID()
		id.PartID = uint32(meta.InternalIdx) + 1000
		id.Idx = meta.ColIdx
		switch meta.IdxType {
		case BlockZoneMapIndex:
			size := idxFile.Stat().Size()
			buf := make([]byte, size)
			if _, err = idxFile.Read(buf); err != nil {
				idxFile.Unref()
				return err
			}
			index.zmReader = NewZMReader(blk.GetBufMgr(), idxFile, id, colDef.Type)
		case StaticFilterIndex:
			size := idxFile.Stat().Size()
			buf := make([]byte, size)
			if _, err = idxFile.Read(buf); err != nil {
				idxFile.Unref()
				return err
			}
			index.bfReader = NewBFReader(blk.GetBufMgr(), idxFile, id)
		default:
			panic("unsupported index type")
		}
	}
	return
}

func (index *immutableIndex) WriteTo(data.Block) error { panic("not supported") }
