// Copyright 2021 Matrix Origin
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package wal

import (
	"context"
	"sync"
	"time"

	"github.com/matrixorigin/matrixone/pkg/vm/engine/tae/logstore/driver/batchstoredriver"
	"github.com/matrixorigin/matrixone/pkg/vm/engine/tae/logstore/entry"
	"github.com/matrixorigin/matrixone/pkg/vm/engine/tae/logstore/sm"
	"github.com/matrixorigin/matrixone/pkg/vm/engine/tae/logstore/store"
)

type entryWithInfo struct {
	e    LogEntry
	info any
}

type DriverConfig struct {
	BatchStoreConfig   *batchstoredriver.StoreCfg
	CheckpointDuration time.Duration
}

type walDriver struct {
	*walInfo
	sync.RWMutex
	impl store.Store
	own  bool

	logInfoQueue sm.Queue

	ckpDuration   time.Duration
	cancelfn      context.CancelFunc
	cancelContext context.Context
	wg            sync.WaitGroup
}

func NewDriver(dir, name string, cfg *DriverConfig) Driver {
	var batchStoreCfg *batchstoredriver.StoreCfg
	ckpDuration := time.Second * 5
	if cfg != nil {
		batchStoreCfg = cfg.BatchStoreConfig
		ckpDuration = cfg.CheckpointDuration
	}
	impl := store.NewStoreWithBatchStoreDriver(dir, name, batchStoreCfg)
	driver := NewDriverWithStore(impl, true, ckpDuration)
	return driver
}

func NewDriverWithStore(impl store.Store, own bool, ckpDuration time.Duration) Driver {
	if ckpDuration == 0 {
		ckpDuration = time.Second
	}
	driver := &walDriver{
		walInfo:     newWalInfo(),
		impl:        impl,
		own:         own,
		wg:          sync.WaitGroup{},
		ckpDuration: ckpDuration,
	}
	driver.logInfoQueue = sm.NewSafeQueue(1000, 1000, driver.onLogInfo)
	driver.cancelContext, driver.cancelfn = context.WithCancel(context.Background())
	driver.logInfoQueue.Start()
	driver.wg.Add(1)
	go driver.checkpointTicker()
	return driver
}

func (driver *walDriver) GetCheckpointed() uint64 {
	return driver.impl.GetCheckpointed(GroupC)
}
func (driver *walDriver) replayhandle(handle store.ApplyHandle) store.ApplyHandle {
	return func(group uint32, commitId uint64, payload []byte, typ uint16, info any) {
		driver.logEntry(info.(*entry.Info))
		handle(group, commitId, payload, typ, nil)
	}
}
func (driver *walDriver) Replay(handle store.ApplyHandle) error {
	return driver.impl.Replay(driver.replayhandle(handle))
}

func (driver *walDriver) GetPenddingCnt() uint64 {
	return driver.impl.GetPendding(GroupC)
}

func (driver *walDriver) GetCurrSeqNum() uint64 {
	return driver.impl.GetCurrSeqNum(GroupC)
}

func (driver *walDriver) LoadEntry(groupID uint32, lsn uint64) (LogEntry, error) {
	return driver.impl.Load(groupID, lsn)
}

func (driver *walDriver) AppendEntry(group uint32, e LogEntry) (uint64, error) {
	id, err := driver.impl.Append(group, e)
	ent := &entryWithInfo{
		e:    e,
		info: e.GetInfo(),
	}
	_, err2 := driver.logInfoQueue.Enqueue(ent)
	if err2 != nil {
		panic(err)
	}
	return id, err
}

func (driver *walDriver) Close() error {
	driver.cancelfn()
	if driver.own {
		return driver.impl.Close()
	}
	driver.logInfoQueue.Stop()
	driver.wg.Wait()
	return nil
}
